unit appcontrol;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, ExtCtrls, translation, machine, Dialogs, LCLType, Controls, appgraphic, Graphics,
  IntfGraphics, renderer, Clipbrd, registryunit, Forms;

type
  TApplicationRendererMode = (rmdNormal, rmdSuccess, rmdError);

  TDetermineReturn = (drAccepted, drRefused, drError);

  TAnimReturnProcOnDone = procedure of object;

  TAppAnimRecord = record
    OnDone: TAnimReturnProcOnDone;
    Animating: Boolean;
    Return: TDetermineReturn;
    Sentence: string;
  end;

  TApplicationControl = class;

  { TApplicationMessageRenderer }

  TApplicationMessageRenderer = class
    private
      Mode: TApplicationRendererMode;
      MessageLabel: string;
      Phase: Byte;
      PhaseInc: Boolean;
      AppImage: TImage;
      App: TApplicationControl;
      PNG: TPortableNetworkGraphic;
      Enable: Boolean;

      procedure Render;
      procedure Draw;
      procedure SetEnabled(const AValue: Boolean);
    public
      constructor Create(Image: TImage; MsgLabel: string; Md: TApplicationRendererMode; A: TApplicationControl);
      procedure SetText(Text: string);
      procedure Tick(var Return: Boolean);

      property Enabled: Boolean read Enable write SetEnabled;
  end;

  TApplicationControlMessages = record
    NoMachine: TApplicationMessageRenderer;
    IncompleteMachine: TApplicationMessageRenderer;
    RemainingSentence: TApplicationMessageRenderer;
  end;

  TApplicationControlScreen = record
    SizeX, SizeY, OffsetX, OffsetY: Integer;
  end;

  { TApplicationControl }

  TApplicationControl = class
    private
      ImageCtrl: TImage; // Objekt Image
      Translate: TApplicationTranslation; // Kniznica prekladu aplikacie
      Machine: TMachine; // Objekt pre automat
      Renderer: TRenderer; // Vlakno casovaca renderovania
      CommonGL: TGraphicLibrary; // Graficka kniznica pre aplikaciu

      SelectedStates: Integer; // Pocet po sebe oznacenych stavov
      SnapToGridSwitch: Boolean; // Prepinac prichytavania stavov k mriezke
      MachineFileName: string; // Adresa a nazov suboru automatu na disku

      Registry: TAppRegistry;
      Path: string;
      ScreenResize: TRendererSync;

      Messages: TApplicationControlMessages; // Renderovane spravy aplikacie
      Screen: TApplicationControlScreen; // Informacie o rozmeroch a ofsete v grafickej ploche
      Animation: TAppAnimRecord; // Informacie o animovani testovania

      function ReadLanguageText(Lbl: string): string; // Vrati text z kniznice prekladu podla zadaneho oznacenia
      procedure RecalculateXY(X,Y: Integer; out Ox, Oy: Integer); // Prepocitava X, Y suradnice pre graficku plochu
      procedure RecalculateClickXY(X,Y: Integer; out Ox, Oy: Integer); // Prepocitava X, Y suradnice pre graficku plochu pre klikani mysou

      procedure RenderGrid; // Vykreslovanie mriezky
      procedure ClearCanvas; // Vymazanie obsahu grafickej plochy
      function LengthOfSet(S: TMachineAlphabet): Integer; // Vrati pocet prvkov v mnozine
      procedure SetSnapToGridValue(const AValue: Boolean);

      procedure AnimationOnDone; // Procedura volana po skonceni animovania
      procedure AnimationSetOffset(Ox, Oy: Integer); // Procedura na nastavenie offsetu pocas animacie
    public
      constructor Create(ImgCtrl: TImage; AppPath: string); // Konstruktor :)
      destructor Destroy; override; // Destruktor :)
      procedure SetAppLanguage(NewLanguage: string); // Nastavuje jazyk aplikacie
      function LoadGraphicLibraries: Boolean; // Nacita graficku kniznicu
      procedure StartRenderer; // Spusti renderovanie
      function OnClose(Form: TForm): Boolean; // Vrati true, ak je mozne ukoncit aplikaciu
      procedure OnOpen(Form: TForm); // Aktualizuje parametre formulara pri starte aplikacie

      procedure CallResizeScreen; // Zavola cez renderer v dalsom ticku proceduru zmeny rozmerov okna
      procedure SetResizeProcedure(ResizeProc: TRendererSync); // Nastavi proceduru na zmenu rozmerov okna

      function CreateNewMachine(Alphabet: TMachineAlphabet): Boolean; // Vytvori novy automat
      function CloseMachine(ForceClose: Boolean=false): Boolean; // Zatvori automat
      function MachineExists: Boolean; // Vrati true, ak automat existuje
      function SaveAs(Dlg: TSaveDialog): Boolean; // Ulozi automat do suboru
      function Save(Dlg: TSaveDialog): Boolean; // Ulozi automat do suboru
      function Open(Dlg: TOpenDialog): Boolean; // Otvori automat so suboru
      function OpenFileFromAppParameter(FileName: string): Boolean; // Otvori subor predany parametrom aplikacii

      procedure SetScreenAndOffset(ScrX, ScrY, OffsetX, OffsetY: Integer); // Nastavuje interne rozmer grafickej plochy a ofset posunutia

      procedure MachineCreateState(X,Y: Integer); // Vytvori novy stav automatu
      procedure SelectObjectAtXY(X, Y: Integer); // Oznaci objekt na pozicii X, Y
      procedure MoveStateToXY(X, Y: Integer); // Presunie stav do pozicie X, Y
      procedure MakeNewStateEvent(SkipElse: Boolean=false); // Vytvori novu udalost
      procedure EditStateEvent; // Upravi vybranu udalost
      procedure DeleteSelected; // Vymaze vybrany objekt
      procedure CopySelectedEventSymbols; // Skopiruje do schranky retazec symbolov vyznacenej udalosti
      function IsStateSelected: Boolean; // Vracia informaciu o vybere stavu
      function IsFirstStateSelected: Boolean; // Vracia informaciu o vybere zaciatocneho stavu
      function IsEventSelected: Boolean; // Vracia informaciu o vybere udalosti
      function IsCyclicEventSelected: Boolean; // Vrati informaciu o oznaceni udalosti vracajucej sa do svojho vychodiskoveho stavu
      function IsElseEventSelected: Boolean; // Vrati informaciu o vybere udalosti typu msetElse
      function IsEventLabelDisplaing: Boolean; // Vrati informaciu o zobrazovani stitku symbolov vybranej udalosti
      procedure SelectNone; // Zrusi vyber;
      procedure SetFirstStateArrowAngle(Angle: Integer); // Nastavy uhol sklonu sipky oznacujucej prvy stav
      procedure SetSelectedStateAsFirst; // Nastavi oznaceny stav ako zaciatocny
      procedure SetStateAsNormal; // Nastavi typ oznaceneho stavu na normalny stav
      procedure SetStateAsAccepting; // Nastavi typ oznaceneho stavu na akceptacny stav
      procedure SetStateAsAcceptingAndEnding; // Nastavi typp oznaceneho stavu na akceptacny a konecny stav
      procedure SetStateAsFailedAndEnding; // Nastavi typ oznaceneho stavu na neuspesny a konecny stav
      procedure SetCyclicEventAngle(Angle: Integer); // Nastavi uhol otocenia kruznice zobrazujucej cyklicku udalost
      procedure SetCyclicEventRadius(Radius: Integer); // Nastavi polomer kruznici cyklickej udalosti
      procedure SetEventLabelDisplaing(Enabled: Boolean); // Nastavi vybranej udalosti ci sa ma alebo nema zorazovat jej stitok
      function GetAvailableSymbolsForEdit: TMachineAlphabet; // Vrati vsetky dostupne symboly pouzitelne pre editaciu vybranej udalosti
      function ReturnSelEventSymbols: string; // Vrati vsetky symboly vybranej udalosti ako retazec

      function TestSentence: TDetermineReturn; // Spusti vyhodnotenie vety automatom a vrati vysledok
      procedure TestSentenceAnimated(ReturnProc: TAnimReturnProcOnDone); // Spusti animovane vyhodnotenie vety automatom
      procedure ForceStopAnimation; // Vynuti zastavenie animovania

      procedure RenderAll; // Vykresli vsetky objekty

      procedure ForceRenderNextTime; // Prikaze vykreslovaciemu vlaknu v dalsom kontrolnom ktoru vyrenderovat obsah plochy

      property Trans[Lbl: String]: string read ReadLanguageText; // Vrati text prekladu na zaklade zadaneho oznaceia
      property GetRenderer: TRenderer read Renderer; // Vrati referenciu na vlakno renderovania
      property SnapToGrid: Boolean read SnapToGridSwitch write SetSnapToGridValue; // Prepina uchytenie k mriezke
      property FileName: string read MachineFileName; // Nazov suboru kam sa uklada automat
      property GetGraphic: TGraphicLibrary read CommonGL; // Vrati odkaz na graficku kniznicu
      property AnimTestResult: TDetermineReturn read Animation.Return; // Vrati vysledok animovaneho testu
      property Animating: Boolean read Animation.Animating; // Vrati informaciu o prebiehajucej animacii
  end;

implementation

{ TApplicationMessageRenderer }

procedure TApplicationMessageRenderer.Render;
var
  BmpText, BmpMask: TBitmap;
  BmpTextIntf, BmpMaskIntf: TLazIntfImage;
  PNGIntf: TLazIntfImage;
  Text: string;
  X,Y: Integer;
  LinePNG, LineText, LineMask: PRGBAColorLine;
  R,G,B: Byte;
begin
  if Enabled = false then Exit;
  Text := App.Trans[MessageLabel];
  R := 0; G := 0; B := 0;
  if Mode = rmdNormal then
    B := 255
  else if Mode = rmdSuccess then
    G := 192
  else
    R := 255;
  BmpText := RenderTextBitmap(Text,R,G,B,255,255,255,26,[fsBold]);
  BmpMask := RenderTextBitmap(Text,Phase,Phase,Phase,0,0,0,26,[fsBold]);

  PNG.PixelFormat := pf32bit;
  PNG.Width := BmpText.Width;
  PNG.Height := BmpText.Height;
  PNG.Canvas.Brush.Color := clBlack;
  PNG.Canvas.FillRect(0,0,PNG.Width,PNG.Height);

  PNGIntf := PNG.CreateIntfImage;
  BmpTextIntf := BmpText.CreateIntfImage;
  BmpMaskIntf := BmpMask.CreateIntfImage;

  for Y := 0 to PNGIntf.Height - 1 do
  begin
    LinePNG := PNGIntf.GetDataLineStart(Y);
    LineText := BmpTextIntf.GetDataLineStart(Y);
    LineMask := BmpMaskIntf.GetDataLineStart(Y);
    for X := 0 to PNGIntf.Width - 1 do
    begin
      LinePNG^[X].Alpha := LineMask^[X].Red;
      LinePNG^[X].Red := LineText^[X].Red;
      LinePNG^[X].Green := LineText^[X].Green;
      LinePNG^[X].Blue := LineText^[X].Blue;
    end;
  end;

  PNG.LoadFromIntfImage(PNGIntf);

  App.GetRenderer.Sync(@Self.Draw);
  //Self.Draw;

  BmpText.Free;
  BmpMask.Free;
  PNGIntf.Free;
  BmpTextIntf.Free;
  BmpMaskIntf.Free;
end;

procedure TApplicationMessageRenderer.Draw;
begin
  AppImage.Canvas.Draw(0,0,PNG);
end;

procedure TApplicationMessageRenderer.SetEnabled(const AValue: Boolean);
Var
  FireTime: TTimeStamp;
begin
  if Enable=AValue then exit;
    Enable:=AValue;

  if Enable = true then
  begin
    Phase := 255;
    PhaseInc := False;
    FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 40);
    App.GetRenderer.AddAction(Self, @Tick, FireTime);
  end
  else
  begin
    //App.GetRenderer.RemoveTickOf(Self);
    App.GetRenderer.ForceRender;
  end;
end;

constructor TApplicationMessageRenderer.Create(Image: TImage; MsgLabel: string; Md: TApplicationRendererMode; A: TApplicationControl);
begin
  MessageLabel := MsgLabel;
  Mode := Md;
  Phase := 255;
  PhaseInc := false;
  App := A;
  PNG := TPortableNetworkGraphic.Create;
  AppImage := Image;
  Enabled := false;
end;

procedure TApplicationMessageRenderer.SetText(Text: string);
begin
  MessageLabel := Text;
end;

procedure TApplicationMessageRenderer.Tick(var Return: Boolean);
Var
  FireTime: TTimeStamp;
begin
  Return := false;
  if Enabled = false then Exit;

  if PhaseInc then
    Inc(Phase, 4)
  else
    Dec(Phase, 4);
  if Phase = 255 then
    PhaseInc := false
  else if Phase <= 64 then
    PhaseInc := true;

  Return := true;

  FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 40);
  App.GetRenderer.AddAction(Self, @Tick, FireTime);
end;

{ TApplicationControl }

function TApplicationControl.ReadLanguageText(Lbl: string): string;
begin
  Result := Translate[Lbl];
end;

procedure TApplicationControl.RecalculateXY(X, Y: Integer; out Ox, Oy: Integer);
begin
  Ox := round(X - (Screen.SizeX / 2) + Screen.OffsetX);
  Oy := round(Y - (Screen.SizeY / 2) + Screen.OffsetY);
end;

procedure TApplicationControl.RecalculateClickXY(X, Y: Integer; out Ox,
  Oy: Integer);
begin
  Ox := X - (Screen.SizeX div 2) - Screen.OffsetX;
  Oy := Y - (Screen.SizeY div 2) - Screen.OffsetY;
end;

procedure TApplicationControl.RenderGrid;
var
  Gx, Gy, I: Integer;
  PC: TColor;
  PW: Integer;
  Color: TColor;
begin
  Gx := Screen.SizeX div 2 - (Screen.SizeX div 2 div 20 + 1) * 20 + Screen.OffsetX mod 20;
  Gy := Screen.SizeY div 2 - (Screen.SizeY div 2 div 20 + 1) * 20 + Screen.OffsetY mod 20;
  PC := ImageCtrl.Canvas.Pen.Color;
  PW := ImageCtrl.Canvas.Pen.Width;
  ImageCtrl.Canvas.Pen.Color := clSilver;
  ImageCtrl.Canvas.Pen.Width := 1;
  Color := RGBToColor(222, 222, 222);
  for I := 0 to Screen.SizeX div 20 + 2 do
  begin
    if Gx + 20 * I = Screen.SizeX div 2 + Screen.OffsetX then
      ImageCtrl.Canvas.Pen.Color := clBlack
    else
      if (Gx + 20 * I - Screen.SizeX div 2 - Screen.OffsetX) mod 200 = 0 then
        ImageCtrl.Canvas.Pen.Color := clSilver
      else
        ImageCtrl.Canvas.Pen.Color := Color;
    ImageCtrl.Canvas.MoveTo(Gx + 20 * I, 0);
    ImageCtrl.Canvas.LineTo(Gx + 20 * I, Screen.SizeY);
  end;
  for I := 0 to Screen.SizeY div 20 + 2 do
  begin
    if Gy + 20 * I = Screen.SizeY div 2 + Screen.OffsetY then
      ImageCtrl.Canvas.Pen.Color := clBlack
    else
      if (Gy + 20 * I - Screen.SizeY div 2 - Screen.OffsetY) mod 200 = 0 then
        ImageCtrl.Canvas.Pen.Color := clSilver
      else
        ImageCtrl.Canvas.Pen.Color := Color;
    ImageCtrl.Canvas.MoveTo(0, Gy + 20 * I);
    ImageCtrl.Canvas.LineTo(Screen.SizeX, Gy + 20 * I);
  end;

  ImageCtrl.Canvas.Pen.Color := clBlack;
  ImageCtrl.Canvas.MoveTo(Screen.SizeX div 2 + Screen.OffsetX, 0);
  ImageCtrl.Canvas.LineTo(Screen.SizeX div 2 + Screen.OffsetX, Screen.SizeY);
  ImageCtrl.Canvas.MoveTo(0, Screen.SizeY div 2 + Screen.OffsetY);
  ImageCtrl.Canvas.LineTo(Screen.SizeX, Screen.SizeY div 2 + Screen.OffsetY);

  ImageCtrl.Canvas.Pen.Color := PC;
  ImageCtrl.Canvas.Pen.Width := PW;
end;

procedure TApplicationControl.ClearCanvas;
begin
  ImageCtrl.Canvas.Brush.Color := clWhite;
  ImageCtrl.Canvas.FillRect(ImageCtrl.ClientRect);
end;

function TApplicationControl.LengthOfSet(S: TMachineAlphabet): Integer;
var
  I: Char;
begin
  Result := 0;
  for I := chr(0) to chr(255) do
    if I in S then
      Inc(Result);
end;

procedure TApplicationControl.SetSnapToGridValue(const AValue: Boolean);
begin
  if SnapToGridSwitch=AValue then exit;
  SnapToGridSwitch:=AValue;

  Registry.WriteBooleanRecord('snaptogrid', AValue);
end;

procedure TApplicationControl.AnimationOnDone;
begin
  Animation.Animating := false;
  if Assigned(Animation.OnDone) then
    Animation.OnDone;
  Messages.RemainingSentence.Enabled := False;
end;

procedure TApplicationControl.AnimationSetOffset(Ox, Oy: Integer);
begin
  Screen.OffsetX := - Ox;
  Screen.OffsetY := - Oy;
end;

constructor TApplicationControl.Create(ImgCtrl: TImage; AppPath: string);
begin
  Path := AppPath;

  ImageCtrl := ImgCtrl;
  Registry := TAppRegistry.Create;

  Translate := TApplicationTranslation.Create(Path);
  Translate.LoadLanguage(Registry.ReadStringRecord('language'));

  Machine := nil;

  Renderer := TRenderer.Create(Self);

  Messages.NoMachine := TApplicationMessageRenderer.Create(ImgCtrl, 'msg_error_no_machine', rmdError, Self);
  Messages.NoMachine.Enabled := True;

  Messages.IncompleteMachine := TApplicationMessageRenderer.Create(ImgCtrl, 'msg_error_machine_incomplete', rmdError, Self);
  Messages.RemainingSentence := TApplicationMessageRenderer.Create(ImgCtrl, '  ', rmdNormal, Self);

  SelectedStates := 0;
  SnapToGridSwitch := Registry.ReadBooleanRecord('snaptogrid');
end;

destructor TApplicationControl.Destroy;
begin
  if not Renderer.Suspended then
  begin
    Renderer.Terminate;
    Renderer.WaitFor;
  end;
  Renderer.Free;
  Renderer := nil;
  Translate.Free;
  Translate := nil;
  Machine.Free;
  Machine := nil;
  ImageCtrl := nil;
  FreeAndNil(Messages.NoMachine);
  FreeAndNil(Messages.IncompleteMachine);
  FreeAndNil(Registry);
  inherited Destroy;
end;

procedure TApplicationControl.SetAppLanguage(NewLanguage: string);
begin
  Registry.WriteStringRecord('language', NewLanguage);
  Translate.LoadLanguage(NewLanguage);
end;

function TApplicationControl.LoadGraphicLibraries: Boolean;
begin
  CommonGL := TGraphicLibrary.Create(Translate, Path);
  Result := CommonGL.LoadLibraryConfig('graphics\common.glc');
end;

procedure TApplicationControl.StartRenderer;
begin
  Renderer.Resume;
end;

function TApplicationControl.OnClose(Form: TForm): Boolean;
var
  Answr: Integer;
  Dlg: TSaveDialog;
begin
  if Animating then
  begin
    Result := false;
    Exit;
  end;

  Result := true;

  if MachineExists then
  begin
    Answr := MessageDlg(Trans['appformcaption'], Trans['appformcloseprompt'], mtConfirmation, mbYesNoCancel, 0);

    if Answr = mrCancel then
    begin
      Result := false;
      Exit;
    end;

    if Answr = mrYes then
    begin
      Dlg := TSaveDialog.Create(nil);
      Result := Save(Dlg);
      FreeAndNil(Dlg);
      if not Result then Exit;
    end;
  end;

  Registry.WriteIntegerRecord('window_x', Form.Left);
  Registry.WriteIntegerRecord('window_y', Form.Top);
  Registry.WriteIntegerRecord('window_width', Form.Width);
  Registry.WriteIntegerRecord('window_height', Form.Height);
  if Form.WindowState = wsMaximized then
    Registry.WriteBooleanRecord('window_maximized', true)
  else
    Registry.WriteBooleanRecord('window_maximized', false);
end;

procedure TApplicationControl.OnOpen(Form: TForm);
var
  Top,Left,Width,Height: Integer;
begin
  if not Registry.FirstUse then
    if Registry.ReadBooleanRecord('window_maximized') then
    begin
      Form.WindowState := wsMaximized;
      Form.Top := Registry.ReadIntegerRecord('window_y');
      Form.Left := Registry.ReadIntegerRecord('window_x');
    end
    else
    begin
      Top := Registry.ReadIntegerRecord('window_y');
      Left := Registry.ReadIntegerRecord('window_x');
      Width := Registry.ReadIntegerRecord('window_width');
      Height := Registry.ReadIntegerRecord('window_height');
      if not ((Top = 0) and (Height = 0) and (Left = 0) and (Width = 0)) then
      begin
        Form.WindowState := wsNormal;
        Form.Top := Top;
        Form.Left := Left;
        Form.Width := Width;
        Form.Height := Height;
      end;
    end;
end;

procedure TApplicationControl.CallResizeScreen;
begin
  GetRenderer.CallScreenResize;
end;

procedure TApplicationControl.SetResizeProcedure(ResizeProc: TRendererSync);
begin
  GetRenderer.SetScreenResizeProcedure(ResizeProc);
  ScreenResize := ResizeProc;
end;

function TApplicationControl.CreateNewMachine(Alphabet: TMachineAlphabet): Boolean;
var
  CreateNew: Boolean;
  Answer: Integer;
begin
  Result := false;
  if Animating then Exit;

  CreateNew := False;

  if Alphabet = [] then
  begin
    MessageDlg(Trans['newmachine_error_caption'], Trans['newmachine_error_message'], mtError, [mbOk], 0, mbOk);
    Exit;
  end
  else
  begin
    CreateNew := True;
  end;

  if Machine <> nil then
  begin
    Answer := MessageDlg(Trans['newmachine_query_caption'], Trans['newmachine_query_question'], mtConfirmation, [mbYes, mbNo], 0, mbYes);
    if Answer = mrYes then
      CreateNew := True
    else
      CreateNew := False;
  end;

  if CreateNew then
  begin
    {if Machine <> nil then
    begin
      Machine.Free;
      Machine := nil;
    end;}
    if MachineExists then
    begin
      if not CloseMachine then Exit;
    end;
    MachineFileName := '';
    Machine := TMachine.Create(ImageCtrl, Alphabet, CommonGL);
    Result := true;
    Messages.NoMachine.Enabled := False;
    Messages.IncompleteMachine.Enabled := True;
    GetRenderer.ForceRender;
  end;
end;

function TApplicationControl.CloseMachine(ForceClose: Boolean=false): Boolean ;
var
  Answr: Integer;
  Dlg: TSaveDialog;
begin
  Result := false;
  if Animating then Exit;

  Result := true;
  if Machine <> nil then
  begin
    if not ForceClose then
    begin
      Answr := MessageDlg(Trans['machine_close_caption'], Trans['machine_close_prompt'], mtConfirmation, mbYesNoCancel, 0);

      if Answr = mrYes then
      begin
        Dlg := TSaveDialog.Create(nil);
        Result := Save(Dlg);
        FreeAndNil(Dlg);
        if not Result then Exit;
      end
      else if Answr = mrCancel then
      begin
        Result := False;
        Exit;
      end;
    end;

    Machine.Free;
    Machine := nil;
    Messages.NoMachine.Enabled := true;
    Messages.IncompleteMachine.Enabled := false;
    MachineFileName := '';
    SelectedStates := 0;
  end;
end;

function TApplicationControl.MachineExists: Boolean;
begin
  Result := Machine <> nil;
end;

function TApplicationControl.SaveAs(Dlg: TSaveDialog): Boolean;
var
  CanSave: Boolean;
begin
  Result := false;
  if Animating then Exit;

  if not MachineExists then Exit;

  Dlg.Title := Trans['files_save_dialog_title'];
  Dlg.DefaultExt := 'mch';
  Dlg.Filter := Trans['files_filter_machine_extension_description'] + '|*.mch|' + Trans['files_filter_all_files_ext_description'] + '|*.*';
  if Dlg.Execute then
  begin
    CanSave := true;

    if FileExists(Dlg.FileName) then
    begin
      CanSave := false;
      if MessageDlg(Trans['files_save_dialog_title'], Trans['files_replace_file_prompt'], mtConfirmation, mbYesNo, 0) = mrYes then
        CanSave := true;
    end;

    if CanSave then
    begin
      Result := Machine.SaveToFile(Dlg.FileName, Translate);
      if Result then MachineFileName := Dlg.FileName;
    end;
  end;
end;

function TApplicationControl.Save(Dlg: TSaveDialog): Boolean;
begin
  Result := False;
  if Animating then Exit;

  if not MachineExists then Exit;

  if MachineFileName = '' then
  begin
    Result := SaveAs(Dlg);
    Exit;
  end;

  Result := Machine.SaveToFile(MachineFileName, Translate);
end;

function TApplicationControl.Open(Dlg: TOpenDialog): Boolean;
begin
  Result := false;
  if Animating then Exit;

  Dlg.Title := Trans['files_open_dialog_title'];
  Dlg.DefaultExt := 'mch';
  Dlg.Filter := Trans['files_filter_machine_extension_description'] + '|*.mch|' + Trans['files_filter_all_files_ext_description'] + '|*.*';
  if Dlg.Execute then
  begin
    if CloseMachine then
    begin
      Machine := TMachine.Create(ImageCtrl, CommonGL);

      Result := Machine.LoadFromFile(Dlg.FileName, Translate);
      if Result then
      begin
        MachineFileName := Dlg.FileName;
        Messages.NoMachine.Enabled := False;
        Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
      end
      else
      begin
        MachineFileName := '';
        CloseMachine(True);
      end;
    end;
  end;
end;

function TApplicationControl.OpenFileFromAppParameter(FileName: string): Boolean;

  function StripQuotes(FileName: string): string;
  begin
    Result := FileName;
    if (FileName[1] = '"') and (FileName[Length(FileName)] = '"') then
      Result := StripQuotes(Copy(FileName, 2, Length(FileName) - 2));
  end;

begin
  Result := true;
  if Animating then Exit;
  if MachineExists then
    CloseMachine(true);

  FileName := StripQuotes(FileName);

  Machine := TMachine.Create(ImageCtrl, CommonGL);
  Result := Machine.LoadFromFile(FileName, Translate);
  if Result then
  begin
    MachineFileName := FileName;
    Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
    Messages.NoMachine.Enabled := false;
  end
  else
  begin
    MachineFileName := '';
    Messages.IncompleteMachine.Enabled := false;
    Messages.NoMachine.Enabled := true;
  end;
end;

procedure TApplicationControl.SetScreenAndOffset(ScrX, ScrY, OffsetX, OffsetY: Integer);
begin
  Screen.SizeX := ScrX;
  Screen.SizeY := ScrY;
  if not Animating then
  begin
    Screen.OffsetX := OffsetX;
    Screen.OffsetY := OffsetY;
  end;
end;

procedure TApplicationControl.MachineCreateState(X, Y: Integer);
var
  Nx, Ny: Integer;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  RecalculateClickXY(X, Y, Nx, Ny);

  Machine.CreateState(Nx, Ny, SnapToGridSwitch);

  Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
end;

procedure TApplicationControl.SelectObjectAtXY(X, Y: Integer);
var
  Nx, Ny: Integer;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  RecalculateClickXY(X, Y, Nx, Ny);

  if Machine.TestSelection(Nx, Ny) then
    Inc(SelectedStates)
  else
    SelectedStates := 0;
end;

procedure TApplicationControl.MoveStateToXY(X, Y: Integer);
var
  Nx, Ny: Integer;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  RecalculateClickXY(X, Y, Nx, Ny);

  Machine.SetXYOfSelState(Nx, Ny, SnapToGridSwitch);
end;

procedure TApplicationControl.MakeNewStateEvent(SkipElse: Boolean=false);
var
  AlphaAvail: TMachineAlphabet;
  I: Char;
  AlphaStr, Output: string;
  E: Integer;
  NewSet: TMachineAlphabet;
  ElsePossible: Boolean;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  if SelectedStates >= 2 then
  begin
    if not Machine.TestEventBetween then
    begin
      MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_exist_error'],mtError,[mbOK],'');
      Machine.SelectNone;
      SelectedStates := 0;
      Exit;
    end;

    AlphaAvail := Machine.GetAvailableSymbolsForEvent;

    ElsePossible := Machine.TestElseEventPossibility;

    if (AlphaAvail = []) or (not ElsePossible and (LengthOfSet(AlphaAvail) = 1)) then
    begin
      MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_no_symbols_available_error'],mtError,[mbOK],'');
      Machine.SelectNone;
      SelectedStates := 0;
      Exit;
    end;

    if ElsePossible and not SkipElse then
    begin
      if MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_create_else'],mtConfirmation,[mbYes, mbNo],'') = mrYes then
      begin
        Machine.CreateStateEvent('', true);
        SelectedStates := 0;
        GetRenderer.ForceRender;
        Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
        Exit;
      end;
    end;

    AlphaStr := '';
    for I := chr(0) to chr(255) do
      if I in AlphaAvail then
        AlphaStr := AlphaStr + I;

    Output := '';
    if InputQuery(Trans['new_state_event_caption'], Trans['new_state_event_prompt'] + #13#13 + AlphaStr, Output) then
    begin
      if Length(Output) = 0 then
      begin
        MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_input_empty_error'],mtError,[mbOK],'');
        MakeNewStateEvent(true);
        Exit;
      end;
      NewSet := [];
      for E := 1 to Length(Output) do
        if not (Output[E] in AlphaAvail) then
        begin
          MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_input_invalid_error'],mtError,[mbOK],'');
          MakeNewStateEvent(true);
          Exit;
        end
        else
          NewSet := NewSet + [Output[E]];
      if not ElsePossible and (NewSet = AlphaAvail) then
      begin
        MessageDlg(Trans['new_state_event_caption'],Trans['new_state_event_input_invalid_else_error'],mtError,[mbOK],'');
        MakeNewStateEvent(true);
        Exit;
      end;
      Machine.CreateStateEvent(Output);
      SelectedStates := 0;
      GetRenderer.ForceRender;
      Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
    end
    else
    begin
      Machine.SelectNone;
      SelectedStates := 0;
    end;
  end;
end;

procedure TApplicationControl.EditStateEvent;
var
  Output: string;
  Symbols, NewSymbols: set of Char;
  SymbStr: string;
  I: Char;
  E: Integer;
  HasElse: Boolean;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Symbols := GetAvailableSymbolsForEdit;

  SymbStr := '';
  for I := #0 to #255 do
    if I in Symbols then
      SymbStr := SymbStr + I;

  Output := ReturnSelEventSymbols;

  HasElse := Machine.TestMyStateHasElseEvent;

  if (Symbols = []) or (HasElse and (LengthOfSet(Symbols) = 1)) then
  begin
    MessageDlg(Trans['edit_state_event_caption'], Trans['edit_state_event_error_no_symbols_available'], mtError, [mbOk], 0);
    EditStateEvent;
    Exit;
  end;

  if InputQuery(Trans['edit_state_event_caption'], Trans['edit_state_event_prompt'] + #13#13 + SymbStr, Output) then
  begin
    NewSymbols := [];
    if Length(Output) = 0 then
    begin
      MessageDlg(Trans['edit_state_event_caption'], Trans['edit_state_event_error_empty'], mtError, [mbOk], 0);
      EditStateEvent;
      Exit;
    end;

    for E := 1 to Length(Output) do
      if not (Output[E] in Symbols) then
      begin
        MessageDlg(Trans['edit_state_event_caption'], Trans['edit_state_event_error_invalid_characters'], mtError, [mbOk], 0);
        EditStateEvent;
        Exit;
      end
      else
        NewSymbols := NewSymbols + [Output[E]];

    if HasElse and (NewSymbols = Symbols) then
    begin
      MessageDlg(Trans['edit_state_event_caption'], Trans['edit_state_event_error_cant_use_all'], mtError, [mbOk], 0);
      EditStateEvent;
      Exit;
    end;

    Machine.SetSelEventNewSymbols(NewSymbols);
  end;
end;

procedure TApplicationControl.DeleteSelected;
var
  Question: string;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  if Machine.IsStateSelected or Machine.IsEventSelected then
  begin
    if Machine.IsStateSelected then
      Question := Trans['delete_state_prompt']
    else
      Question := Trans['delete_event_prompt'];

    if MessageDlg(Trans['delete_object_caption'], Question, mtConfirmation, mbYesNo, 0) = mrNo then
      Exit;
  end
  else
    Exit;

  Machine.DeleteSelected;

  SelectedStates := 0;

  Messages.IncompleteMachine.Enabled := not Machine.TestIfMachineIsComplete;
end;

procedure TApplicationControl.CopySelectedEventSymbols;
var
  CB: TClipboard;
begin
  if Animating then Exit;
  CB := TClipboard.Create;
  CB.Clear;
  if MachineExists then
    CB.AsText := Machine.ReturnSelEventSymbols;
  CB.Free;
end;

function TApplicationControl.IsStateSelected: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsStateSelected;
end;

function TApplicationControl.IsFirstStateSelected: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsFirstStateSelected;
end;

function TApplicationControl.IsEventSelected: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsEventSelected;
end;

function TApplicationControl.IsCyclicEventSelected: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsCyclicEventSelected;
end;

function TApplicationControl.IsElseEventSelected: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsElseEventSelected;
end;

function TApplicationControl.IsEventLabelDisplaing: Boolean;
begin
  Result := false;

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.IsEventLabelDisplaing;
end;

procedure TApplicationControl.SelectNone;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SelectNone;

  SelectedStates := 0;

  GetRenderer.ForceRender;
end;

procedure TApplicationControl.SetFirstStateArrowAngle(Angle: Integer);
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetFirstStateArrowAngle(Angle);
end;

procedure TApplicationControl.SetSelectedStateAsFirst;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.MakeStateFirst;
end;

procedure TApplicationControl.SetStateAsNormal;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelStateType(mstNormal);
end;

procedure TApplicationControl.SetStateAsAccepting;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelStateType(mstAcceptState);
end;

procedure TApplicationControl.SetStateAsAcceptingAndEnding;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelStateType(mstAcceptAndEndState);
end;

procedure TApplicationControl.SetStateAsFailedAndEnding;
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelStateType(mstFailAndEnd);
end;

procedure TApplicationControl.SetCyclicEventAngle(Angle: Integer);
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelEventAngle(Angle);
end;

procedure TApplicationControl.SetCyclicEventRadius(Radius: Integer);
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelEventRadius(Radius);
end;

procedure TApplicationControl.SetEventLabelDisplaing(Enabled: Boolean);
begin
  if Animating then Exit;
  if not MachineExists then Exit;

  Machine.SetSelEventLabelDisplaing(Enabled);
end;

function TApplicationControl.GetAvailableSymbolsForEdit: TMachineAlphabet;
begin
  Result := [];

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.GetAvailableSymbolsForEdit;
end;

function TApplicationControl.ReturnSelEventSymbols: string;
begin
  Result := '';

  if Animating then Exit;
  if not MachineExists then Exit;

  Result := Machine.ReturnSelEventSymbols;
end;

function TApplicationControl.TestSentence: TDetermineReturn;
var
  Alphabet: TMachineAlphabet;
  Output, Allowed: String;
  E: Char;
  I: Integer;
begin
  Result := drError;
  if Animating then Exit;

  if not MachineExists then
  begin
    MessageDlg(Trans['machine_determine_sentence_caption'], Trans['machine_determine_sentence_error_no_machine'], mtError, [mbOK], 0);
    Exit;
  end;

  if not Machine.TestIfMachineIsComplete then
  begin
    MessageDlg(Trans['machine_determine_sentence_caption'], Trans['machine_determine_sentence_error_incomplete'], mtError, [mbOK], 0);
    Exit;
  end;

  Output := '';

  Alphabet := Machine.GetAlphabet;

  Allowed := '';
  for E := #0 to #255 do
    if E in Alphabet then
      Allowed := Allowed + E;

  if InputQuery(Trans['machine_determine_sentence_caption'], Trans['machine_determine_sentence_prompt'] + #13#13 + Allowed, Output) then
  begin
    for I := 1 to Length(Output) do
      if not (Output[I] in Alphabet) then
      begin
        MessageDlg(Trans['machine_determine_sentence_caption'], Trans['machine_determine_sentence_error_invalid_chars'], mtError, [mbOK], 0);
        Exit;
      end;

    Animation.Sentence := Output;
    if Machine.Determine(Output) then
      Result := drAccepted
    else
      Result := drRefused;
  end;
end;

procedure TApplicationControl.TestSentenceAnimated(ReturnProc: TAnimReturnProcOnDone);
begin
  Animation.Return := TestSentence;
  if Animation.Return <> drError then
  begin
    Animation.Animating := true;
    Animation.OnDone := ReturnProc;
    Machine.StartAnimation(Animation.Sentence, @AnimationOnDone, @AnimationSetOffset, GetRenderer);
    Messages.RemainingSentence.Enabled := true;
  end;
end;

procedure TApplicationControl.ForceStopAnimation;
begin
  if Animating and MachineExists then
    Machine.ForceStopAnimation;
end;

procedure TApplicationControl.RenderAll;
var
  Mx,My: Integer;
begin
  if Assigned(ScreenResize) then
    ScreenResize;

  Self.ClearCanvas;

  RenderGrid;

  if MachineExists then
  begin
    RecalculateXY(Screen.SizeX, Screen.SizeY, Mx, My);
    Machine.Render(Mx, My);

    Machine.NumberOfStatesAndEvents(Mx, My);
  end;

  Messages.NoMachine.Render;
  Messages.IncompleteMachine.Render;
  if MachineExists then
    Messages.RemainingSentence.SetText(Trans['anim_remaining_sentence'] + ': ' + Machine.RemainingSentence);
  Messages.RemainingSentence.Render;
end;

procedure TApplicationControl.ForceRenderNextTime;
begin
  GetRenderer.ForceRender;
end;

end.

