unit registryunit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, registry;

const REGISTRY_KEY_PATH = 'Software\FSM';

type

  { TAppRegistry }

  TAppRegistry = class
    private
      Registry: TRegistry;
      IsFirstUse: Boolean;
    public
      constructor Create;
      destructor Destroy; override;
      function ReadStringRecord(Name: string): string;
      function ReadIntegerRecord(Name: string): Integer;
      function ReadBooleanRecord(Name: string): Boolean;
      procedure WriteStringRecord(Name: string; Value: string);
      procedure WriteIntegerRecord(Name: string; Value: Integer);
      procedure WriteBooleanRecord(Name: string; Value: Boolean);

      property FirstUse: Boolean read IsFirstUse;
  end;

implementation

{ TAppRegistry }

constructor TAppRegistry.Create;
begin
  IsFirstUse := false;
  Registry := TRegistry.Create;
  Registry.RootKey := HKEY_CURRENT_USER;
  if not Registry.KeyExists(REGISTRY_KEY_PATH) then
  begin
    Registry.CreateKey(REGISTRY_KEY_PATH);
    IsFirstUse := true;
  end;
end;

destructor TAppRegistry.Destroy;
begin
  FreeAndNil(Registry);

  inherited Destroy;
end;

function TAppRegistry.ReadStringRecord(Name: string): string;
begin
  try
    Registry.OpenKeyReadOnly(REGISTRY_KEY_PATH);
    Result := Registry.ReadString(Name);
  except
    Result := '';
  end;
  Registry.CloseKey;
end;

function TAppRegistry.ReadIntegerRecord(Name: string): Integer;
begin
  try
    Registry.OpenKeyReadOnly(REGISTRY_KEY_PATH);
    Result := Registry.ReadInteger(Name);
  except
    Result := 0;
  end;
  Registry.CloseKey;
end;

function TAppRegistry.ReadBooleanRecord(Name: string): Boolean;
begin
  try
    Registry.OpenKeyReadOnly(REGISTRY_KEY_PATH);
    Result := Registry.ReadBool(Name);
  except
    Result := false;
  end;
  Registry.CloseKey;
end;

procedure TAppRegistry.WriteStringRecord(Name: string; Value: string);
begin
  try
    Registry.OpenKey(REGISTRY_KEY_PATH, true);
    Registry.WriteString(Name, Value);
  finally
    Registry.CloseKey;
  end;
end;

procedure TAppRegistry.WriteIntegerRecord(Name: string; Value: Integer);
begin
  try
    Registry.OpenKey(REGISTRY_KEY_PATH, true);
    Registry.WriteInteger(Name, Value);
  finally
    Registry.CloseKey;
  end;
end;

procedure TAppRegistry.WriteBooleanRecord(Name: string; Value: Boolean);
begin
  try
    Registry.OpenKey(REGISTRY_KEY_PATH, true);
    Registry.WriteBool(Name, Value);
  finally
    Registry.CloseKey;
  end;
end;

end.

